#include <string.h>
#include <stdlib.h>
#include "Packet.h"
#include "msgpack.h"

enum {
  ERROR_PACKET_PACKETID = 1,
  ERROR_PACKET_TYPE_ASSET = 2,
  ERROR_PACKET_UNKNOWN_PACKETID = 3,
};

static int encode(struct Packet* pkt, msgpack_sbuffer* buffer) {
  msgpack_packer packer;

  msgpack_packer_init(&packer, buffer, msgpack_sbuffer_write);
  int off = msgpack_pack_uint16(&packer, pkt->PacketId);
  printf("return value of msgpack_pack_uint16 = %d\n", off);
  switch (pkt->PacketId) {
    case PLogin: {
      struct LoginPacket* l = (struct LoginPacket*)pkt;
      msgpack_pack_array(&packer, 2);
      msgpack_pack_raw(&packer, l->Username.size);
      printf("sizeof Username is %d\n", l->Username.size);
      msgpack_pack_raw_body(&packer, l->Username.ptr, l->Username.size);
      printf("after pack username:%zu\n", buffer->size);

      msgpack_pack_raw(&packer, l->Password.size);
      msgpack_pack_raw_body(&packer, l->Password.ptr, l->Password.size);
      printf("after pack Password:%zu\n", buffer->size);

      msgpack_unpacked msg;
      msgpack_unpacked_init(&msg);
      size_t off = 4;
      bool success =
          msgpack_unpack_next(&msg, buffer->data, buffer->size, &off);
      if (success) {
        printf("off = %zu\n", off);
      }

      /* prints the deserialized object. */
      msgpack_object obj = msg.data;
      msgpack_object_print(stdout, obj);

      success = msgpack_unpack_next(&msg, buffer->data, buffer->size, &off);

      /* prints the deserialized object. */
      obj = msg.data;
      msgpack_object_print(stdout, obj);
    } break;
    case PSelectCharactor: {
      struct SelectCharactorPacket* p = (struct SelectCharactorPacket*)pkt;
      msgpack_pack_array(&packer, 1);
      msgpack_pack_int(&packer, p->Which);
    } break;
    default:
      return -1;
  }
  return 0;
}

// 打包的结果中buf的大小包括了4个字节的包头部
int Marshal(struct Packet* pkt, msgpack_sbuffer* buf) {
  uint32_t tmp = 0;

  msgpack_sbuffer_clear(buf);
  msgpack_sbuffer_write(buf, (const char*)&tmp, 4); // 留4个字节用来写头部包大小

  encode(pkt, buf);

  printf("write the uint32 size: %zu\n", buf->size - 4);
  WriteU32(buf->data, (uint32_t)buf->size - 4);
  for (int i = 0; i < 4; i++) {
    printf("%x ", buf->data[i]);
  }
  return 0;
}

// 解包时buf中并不包含4个字节的包头，offset从0算起
int Unmarshal(msgpack_sbuffer* buf, struct Packet* pkt) {
  msgpack_unpacked msg;
  msgpack_object obj;
  size_t off;

  off = 0;
  msgpack_unpacked_init(&msg);
  bool succ = msgpack_unpack_next(&msg, buf->data, buf->size - off, &off);
  if (!succ) {
    return ERROR_PACKET_PACKETID;
  }

  if (msg.data.type != MSGPACK_OBJECT_POSITIVE_INTEGER) {
    return ERROR_PACKET_TYPE_ASSET;
  }

  pkt->PacketId = msg.data.via.u64;
  printf("run here....offset= %zu\n", off);

  switch (pkt->PacketId) {
    case PCharactorInfo: {
      struct CharactorInfoPacket* p = (struct CharactorInfoPacket*)pkt;
      bool success =
          msgpack_unpack_next(&msg, buf->data, buf->size - off, &off);
      if (!success) {
        printf("msgpack_unpack_next failed!");
      }

      if (msg.data.type != MSGPACK_OBJECT_ARRAY) {
        return -1;
      }
      obj = msg.data.via.array.ptr[0];
      p->Name = obj.via.raw;
      obj = msg.data.via.array.ptr[1];
      p->Class = obj.via.raw;
      obj = msg.data.via.array.ptr[2];
      p->Level = obj.via.u64;
    } break;
    case PLogin: {
      struct LoginPacket* p = (struct LoginPacket*)pkt;
      printf("run here....offset= %zu\n", off);
      bool success = msgpack_unpack_next(&msg, buf->data, buf->size, &off);
      //            success = msgpack_unpack_next(&msg, buf->data, buf->size,
      // &off);
      if (!success) {
        printf("msgpack_unpack_next failed!");
      }
      printf("........\n");
      msgpack_object_print(stdout, msg.data);
      printf("........\n");

      if (msg.data.type != MSGPACK_OBJECT_ARRAY) {
        return -1;
      }
      obj = msg.data.via.array.ptr[0];
      p->Username = obj.via.raw;
      obj = msg.data.via.array.ptr[1];
      p->Password = obj.via.raw;
    } break;
    case PLoginOk: {
      printf("......get a PLoginOk");
    } break;
    default:
      return ERROR_PACKET_UNKNOWN_PACKETID;
  }
  msgpack_unpacked_destroy(&msg);
  return 0;
}

void expand(struct Slice* buf) {
  char* tmp = (char*)malloc(2 * buf->cap);
  memcpy(tmp, buf->data, buf->cap);
  free(buf->data);
  buf->data = tmp;
  buf->cap *= 2;
}

void WriteString(struct Slice* buf, char* data) {
  unsigned long len = strlen(data);
  while (buf->len + len + 2 > buf->cap) {
    expand(buf);
  }
  WriteU16(buf, (uint16_t)len);
  memcpy(buf->data + buf->len, data, len);
  buf->len += len;
}

void WriteU16(struct Slice* buf, uint16_t data) {
  while (buf->len + 2 > buf->cap) {
    expand(buf);
  }
  buf->data[buf->len] = data >> 8;
  buf->data[buf->len + 1] = data & 255;
  buf->len += 2;
}

void WriteU32(char* buf, uint32_t data) {
  buf[0] = (char)(data >> 24);
  buf[1] = (char)(data >> 16);
  buf[2] = (char)(data >> 8);
  buf[3] = (char)(data);
}

void WriteS32(struct Slice* buf, int32_t data) {
  //   WriteU32(buf, data);
}

int ReadString(struct Slice* buf, char** data) {
  uint16_t size;
  int ret;

  ret = ReadU16(buf, &size);
  if (ret < 0)
    goto error;

  if (buf->pos + size > buf->len) {
    ret = -1;
    goto error;
  }

  *data = (char*)malloc(size);
  memcpy(*data, buf->data + buf->pos, size);
  buf->pos += size;
  return 0;
error:
  return ret;
}

int ReadU32(char* buf, uint32_t* data) {
  *data = 0;
  *data = (*buf) >> 24;
  *data |= (*(buf + 1)) >> 16;
  *data |= (*(buf + 2)) >> 8;
  *data |= (*(buf + 3));

  return 0;
}

int ReadU16(struct Slice* buf, uint16_t* data) {
  if (buf->pos + 2 > buf->len) {
    return -1;
  }

  *data = (*(buf->data + buf->pos)) >> 8;
  *data |= (*(buf->data + buf->pos + 1));

  buf->pos += 2;
  return 0;
}

void InitSlice(struct Slice* buf, int cap) {
  buf->data = (char*)malloc(cap);
  buf->len = 0;
  buf->cap = cap;
}

#ifdef UNIT_TEST
#include <stdio.h>

int main() {
  struct LoginPacket login;
  int ret;
  msgpack_sbuffer buf;

  msgpack_sbuffer_init(&buf);

  login.PacketId = PLogin;
  login.Username.ptr = "genius";
  login.Username.size = 6;
  login.Password.ptr = "0101001";
  login.Password.size = 7;

  ret = Marshal((struct Packet*)&login, &buf);
  if (ret < 0) {
    printf("Marshal error...");
  }

  FILE* f = fopen("data.out", "w");
  if (f != NULL) {
    fwrite(buf.data, buf.size, 1, f);
    fflush(f);
    fclose(f);
  } else {
    printf("打开文件失败");
  }

  struct LoginPacket pkt;
  ret = Unmarshal(&buf, &pkt);
  if (ret < 0) {
    printf("unmarshal return a error: %d\n", ret);
  }

  if (pkt.PacketId != PLogin) {
    printf("get a wrong packet id: %d!!\n", pkt.PacketId);
  }
  printf("Username: %s\n", pkt.Username.ptr);
  printf("Password: %s\n", pkt.Password.ptr);

  msgpack_sbuffer_destroy(&buf);
  return 0;
}
#endif
